netwire setup script
Copyright (C) 2013, Ertugrul Soeylemez

Please see the LICENSE file for terms and conditions of use,
modification and distribution of this package, including this file.

> module Main where
>
> import Distribution.Simple
>
> main :: IO ()
> main = defaultMain
